import json

from mock import MagicMock
from nose.tools import with_setup, assert_equals, raises
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.livyclientlib.exceptions import (
    SessionManagementException,
    HttpClientException,
)
from sparkmagic.livyclientlib.sparkcontroller import SparkController

client_manager = None
controller = None
ipython_display = None


class DummyResponse:
    def __init__(self, status_code, json_text):
        self._status_code = status_code
        self._json_text = json_text

    def json(self):
        return json.loads(self._json_text)

    @property
    def status_code(self):
        return self._status_code


def _setup():
    global client_manager, controller, ipython_display

    client_manager = MagicMock()
    ipython_display = MagicMock()
    spark_events = MagicMock()
    controller = SparkController(ipython_display)
    controller.session_manager = client_manager
    controller.spark_events = spark_events


def _teardown():
    pass


@with_setup(_setup, _teardown)
def test_add_session():
    name = "name"
    properties = {"kind": "spark"}
    endpoint = Endpoint("http://location:port", None)
    session = MagicMock()

    controller._livy_session = MagicMock(return_value=session)
    controller._http_client = MagicMock(return_value=MagicMock())

    controller.add_session(name, endpoint, False, properties)

    controller._livy_session.assert_called_once_with(
        controller._http_client.return_value, properties, ipython_display
    )
    controller.session_manager.add_session.assert_called_once_with(name, session)
    session.start.assert_called_once()


@with_setup(_setup, _teardown)
def test_add_session_skip():
    name = "name"
    language = "python"
    connection_string = "url=http://location:port;username=name;password=word"
    client = "client"
    session = MagicMock()
    controller._livy_session = MagicMock(return_value=session)
    controller._http_client = MagicMock(return_value=client)

    client_manager.get_sessions_list.return_value = [name]
    controller.add_session(name, language, connection_string, True)

    assert controller._livy_session.create_session.call_count == 0
    assert controller._http_client.build_client.call_count == 0
    assert client_manager.add_session.call_count == 0
    assert session.start.call_count == 0


@with_setup(_setup, _teardown)
def test_delete_session():
    name = "name"

    controller.delete_session_by_name(name)

    client_manager.delete_client.assert_called_once_with(name)


@with_setup(_setup, _teardown)
def test_cleanup():
    controller.cleanup()
    client_manager.clean_up_all.assert_called_once_with()


@with_setup(_setup, _teardown)
def test_run_cell():
    default_client = MagicMock()
    chosen_client = MagicMock()
    client_manager.get_any_session = MagicMock(return_value=default_client)
    client_manager.get_session = MagicMock(return_value=chosen_client)
    name = "session_name"
    command = MagicMock()

    controller.run_command(command, name)
    command.execute.assert_called_with(chosen_client)

    controller.run_command(command, None)
    command.execute.assert_called_with(default_client)


@with_setup(_setup, _teardown)
def test_run_sql():
    default_client = MagicMock()
    chosen_client = MagicMock()
    client_manager.get_any_session = MagicMock(return_value=default_client)
    client_manager.get_session = MagicMock(return_value=chosen_client)
    name = "session_name"
    sqlquery = MagicMock()

    controller.run_sqlquery(sqlquery, name)
    sqlquery.execute.assert_called_with(chosen_client)

    controller.run_sqlquery(sqlquery, None)
    sqlquery.execute.assert_called_with(default_client)


@with_setup(_setup, _teardown)
def test_get_client_keys():
    controller.get_client_keys()
    client_manager.get_sessions_list.assert_called_once_with()


@with_setup(_setup, _teardown)
def test_get_all_sessions():
    http_client = MagicMock()
    http_client.get_sessions.return_value = json.loads(
        '{"from":0,"total":3,"sessions":[{"id":0,"state":"idle","kind":'
        '"spark","log":[""]}, {"id":1,"state":"busy","kind":"spark","log"'
        ':[""]},{"id":2,"state":"busy","kind":"sql","log"'
        ':[""]}]}'
    )
    controller._http_client = MagicMock(return_value=http_client)
    controller._livy_session = MagicMock()

    sessions = controller.get_all_sessions_endpoint("conn_str")

    assert len(sessions) == 2


@with_setup(_setup, _teardown)
def test_cleanup_endpoint():
    s0 = MagicMock()
    s1 = MagicMock()
    controller.get_all_sessions_endpoint = MagicMock(return_value=[s0, s1])

    controller.cleanup_endpoint("conn_str")

    s0.delete.assert_called_once_with()
    s1.delete.assert_called_once_with()


@with_setup(_setup, _teardown)
def test_delete_session_by_id_existent_non_managed():
    http_client = MagicMock()
    http_client.get_session.return_value = json.loads(
        '{"id":0,"state":"starting","kind":"spark","log":[]}'
    )
    controller._http_client = MagicMock(return_value=http_client)
    session = MagicMock()
    controller._livy_session = MagicMock(return_value=session)

    controller.delete_session_by_id("conn_str", 0)

    controller._livy_session.assert_called_once_with(
        http_client, {"kind": "spark"}, ipython_display, 0
    )
    session.delete.assert_called_once_with()


@with_setup(_setup, _teardown)
def test_delete_session_by_id_existent_managed():
    name = "name"
    controller.session_manager.get_session_name_by_id_endpoint = MagicMock(
        return_value=name
    )
    controller.session_manager.get_sessions_list = MagicMock(return_value=[name])
    controller.delete_session_by_name = MagicMock()

    controller.delete_session_by_id("conn_str", 0)

    controller.delete_session_by_name.assert_called_once_with(name)


@with_setup(_setup, _teardown)
@raises(HttpClientException)
def test_delete_session_by_id_non_existent():
    http_client = MagicMock()
    http_client.get_session.side_effect = HttpClientException
    controller._http_client = MagicMock(return_value=http_client)
    session = MagicMock()
    controller._livy_session = MagicMock(return_value=session)

    controller.delete_session_by_id("conn_str", 0)


@with_setup(_setup, _teardown)
def test_get_app_id():
    chosen_client = MagicMock()
    controller.get_session_by_name_or_default = MagicMock(return_value=chosen_client)

    result = controller.get_app_id()
    assert_equals(result, chosen_client.get_app_id.return_value)
    chosen_client.get_app_id.assert_called_with()


@with_setup(_setup, _teardown)
def test_get_driver_log():
    chosen_client = MagicMock()
    controller.get_session_by_name_or_default = MagicMock(return_value=chosen_client)

    result = controller.get_driver_log_url()
    assert_equals(result, chosen_client.get_driver_log_url.return_value)
    chosen_client.get_driver_log_url.assert_called_with()


@with_setup(_setup, _teardown)
def test_get_logs():
    chosen_client = MagicMock()
    controller.get_session_by_name_or_default = MagicMock(return_value=chosen_client)

    result = controller.get_logs()
    assert_equals(result, chosen_client.get_logs.return_value)
    chosen_client.get_logs.assert_called_with()


@with_setup(_setup, _teardown)
@raises(SessionManagementException)
def test_get_logs_error():
    chosen_client = MagicMock()
    controller.get_session_by_name_or_default = MagicMock(
        side_effect=SessionManagementException("THERE WAS A SPOOKY GHOST")
    )

    result = controller.get_logs()


@with_setup(_setup, _teardown)
def test_get_session_id_for_client():
    assert controller.get_session_id_for_client("name") is not None
    client_manager.get_session_id_for_client.assert_called_once_with("name")


@with_setup(_setup, _teardown)
def test_get_spark_ui_url():
    chosen_client = MagicMock()
    controller.get_session_by_name_or_default = MagicMock(return_value=chosen_client)

    result = controller.get_spark_ui_url()
    assert_equals(result, chosen_client.get_spark_ui_url.return_value)
    chosen_client.get_spark_ui_url.assert_called_with()


@with_setup(_setup, _teardown)
def test_add_session_throws_when_session_start_fails():
    name = "name"
    properties = {"kind": "spark"}
    endpoint = Endpoint("http://location:port", None)
    session = MagicMock()

    controller._livy_session = MagicMock(return_value=session)
    controller._http_client = MagicMock(return_value=MagicMock())
    e = ValueError("Failed to create the SqlContext.\nError, '{}'".format("Exception"))
    session.start = MagicMock(side_effect=e)

    try:
        controller.add_session(name, endpoint, False, properties)
        assert False
    except ValueError as ex:
        assert str(ex) == str(e)
        session.start.assert_called_once()
        controller.session_manager.add_session.assert_not_called


@with_setup(_setup, _teardown)
def test_add_session_cleanup_when_timeouts_and_session_posted_to_livy():
    pass


@with_setup(_setup, _teardown)
def test_add_session_cleanup_when_timeouts_and_session_posted_to_livy():
    _do_test_add_session_cleanup_when_timeouts(is_session_posted_to_livy=True)


@with_setup(_setup, _teardown)
def test_add_session_cleanup_when_timeouts_and_session_not_posted_to_livy():
    _do_test_add_session_cleanup_when_timeouts(is_session_posted_to_livy=False)


def _do_test_add_session_cleanup_when_timeouts(is_session_posted_to_livy):
    name = "name"
    properties = {"kind": "spark"}
    endpoint = Endpoint("http://location:port", None)
    session = MagicMock()

    controller._livy_session = MagicMock(return_value=session)
    controller._http_client = MagicMock(return_value=MagicMock())
    e = RuntimeError("Time out while post session to livy")
    session.start = MagicMock(side_effect=e)

    if is_session_posted_to_livy:
        session.is_posted = MagicMock(return_value=True)
    else:
        session.is_posted = MagicMock(return_value=False)

    try:
        controller.add_session(name, endpoint, False, properties)
        assert False
    except RuntimeError as ex:
        assert str(ex) == str(e)
        session.start.assert_called_once()
        controller.session_manager.add_session.assert_not_called

        if is_session_posted_to_livy:
            session.delete.assert_called_once()
        else:
            session.delete.assert_not_called()


@with_setup(_setup, _teardown)
def test_add_session_cleanup_when_session_delete_throws():
    name = "name"
    properties = {"kind": "spark"}
    endpoint = Endpoint("http://location:port", None)
    session = MagicMock()

    controller._livy_session = MagicMock(return_value=session)
    controller._http_client = MagicMock(return_value=MagicMock())
    e = RuntimeError("Time out while post session to livy")
    session.start = MagicMock(side_effect=e)
    session.is_posted = MagicMock(return_value=True)

    error_from_cleanup = RuntimeError("Error while clean up session")
    session.delete = MagicMock(side_effect=error_from_cleanup)
    try:
        controller.add_session(name, endpoint, False, properties)
        assert False
    except Exception as ex:
        # in the exception chain mechanism, original exception will be set as context.
        assert str(ex) == str(error_from_cleanup)
        assert str(ex.__context__) == str(e)
        controller.session_manager.add_session.assert_not_called



import sys
from base64 import b64encode
from contextlib import contextmanager
from mock import MagicMock
from nose.tools import assert_equals, with_setup

from IPython.display import Image

import sparkmagic.livyclientlib.exceptions
import sparkmagic.utils.configuration as conf
from sparkmagic.utils.constants import (
    SESSION_KIND_SPARK,
    MIMETYPE_IMAGE_PNG,
    MIMETYPE_TEXT_HTML,
    MIMETYPE_TEXT_PLAIN,
    COMMAND_INTERRUPTED_MSG,
)
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.livysession import LivySession
from sparkmagic.livyclientlib.exceptions import SparkStatementCancelledException
from . import test_livysession as tls


if sys.version_info[0] == 2:
    from StringIO import StringIO
elif sys.version_info[0] == 3:
    from io import StringIO
else:
    assert False


def _setup():
    conf.override_all({})


def _create_session(
    kind=SESSION_KIND_SPARK, session_id=-1, http_client=None, spark_events=None
):
    if http_client is None:
        http_client = MagicMock()
    if spark_events is None:
        spark_events = MagicMock()
    ipython_display = MagicMock()
    session = LivySession(
        http_client,
        {"kind": kind, "heartbeatTimeoutInSecond": 60},
        ipython_display,
        session_id,
        spark_events,
    )
    return session


@contextmanager
def _capture_stderr():
    try:
        sys.stderr = StringIO()
        yield sys.stderr
    finally:
        sys.stderr = sys.__stderr__


@with_setup(_setup)
def test_execute():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.post_session.return_value = tls.TestLivySession.session_create_json
    http_client.post_statement.return_value = tls.TestLivySession.post_statement_json
    http_client.get_session.return_value = tls.TestLivySession.ready_sessions_json
    http_client.get_statement.return_value = tls.TestLivySession.ready_statement_json
    session = _create_session(kind=kind, http_client=http_client)
    session.start()
    command = Command("command", spark_events=spark_events)

    result = command.execute(session)

    http_client.post_statement.assert_called_with(0, {"code": command.code})
    http_client.get_statement.assert_called_with(0, 0)
    assert result[0]
    assert_equals(tls.TestLivySession.pi_result, result[1])
    assert_equals(MIMETYPE_TEXT_PLAIN, result[2])
    spark_events.emit_statement_execution_start_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid
    )
    spark_events.emit_statement_execution_end_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid, 0, True, "", ""
    )

    # Now try with PNG result:
    http_client.get_statement.return_value = {
        "id": 0,
        "state": "available",
        "output": {
            "status": "ok",
            "execution_count": 0,
            "data": {"text/plain": "", "image/png": b64encode(b"hello")},
        },
    }
    result = command.execute(session)
    assert result[0]
    assert isinstance(result[1], Image)
    assert result[1].data == b"hello"
    assert_equals(MIMETYPE_IMAGE_PNG, result[2])

    # Now try with HTML result:
    http_client.get_statement.return_value = {
        "id": 0,
        "state": "available",
        "output": {
            "status": "ok",
            "execution_count": 0,
            "data": {"text/html": "<p>out</p>"},
        },
    }
    result = command.execute(session)
    assert result[0]
    assert_equals("<p>out</p>", result[1])
    assert_equals(MIMETYPE_TEXT_HTML, result[2])


@with_setup(_setup)
def test_execute_waiting():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.post_session.return_value = tls.TestLivySession.session_create_json
    http_client.post_statement.return_value = tls.TestLivySession.post_statement_json
    http_client.get_session.return_value = tls.TestLivySession.ready_sessions_json
    http_client.get_statement.side_effect = [
        tls.TestLivySession.waiting_statement_json,
        tls.TestLivySession.waiting_statement_json,
        tls.TestLivySession.ready_statement_json,
        tls.TestLivySession.ready_statement_json,
    ]
    session = _create_session(kind=kind, http_client=http_client)
    session.start()
    command = Command("command", spark_events=spark_events)

    result = command.execute(session)

    http_client.post_statement.assert_called_with(0, {"code": command.code})
    http_client.get_statement.assert_called_with(0, 0)
    assert result[0]
    assert_equals(tls.TestLivySession.pi_result, result[1])
    assert_equals(MIMETYPE_TEXT_PLAIN, result[2])
    spark_events.emit_statement_execution_start_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid
    )
    spark_events.emit_statement_execution_end_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid, 0, True, "", ""
    )


@with_setup(_setup)
def test_execute_null_ouput():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.post_session.return_value = tls.TestLivySession.session_create_json
    http_client.post_statement.return_value = tls.TestLivySession.post_statement_json
    http_client.get_session.return_value = tls.TestLivySession.ready_sessions_json
    http_client.get_statement.return_value = (
        tls.TestLivySession.ready_statement_null_output_json
    )
    session = _create_session(kind=kind, http_client=http_client)
    session.start()
    command = Command("command", spark_events=spark_events)

    result = command.execute(session)

    http_client.post_statement.assert_called_with(0, {"code": command.code})
    http_client.get_statement.assert_called_with(0, 0)
    assert result[0]
    assert_equals("", result[1])
    assert_equals(MIMETYPE_TEXT_PLAIN, result[2])
    spark_events.emit_statement_execution_start_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid
    )
    spark_events.emit_statement_execution_end_event.assert_called_once_with(
        session.guid, session.kind, session.id, command.guid, 0, True, "", ""
    )


@with_setup(_setup)
def test_execute_failure_wait_for_session_emits_event():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.post_session.return_value = tls.TestLivySession.session_create_json
    http_client.post_statement.return_value = tls.TestLivySession.post_statement_json
    http_client.get_session.return_value = tls.TestLivySession.ready_sessions_json
    http_client.get_statement.return_value = tls.TestLivySession.ready_statement_json
    session = _create_session(kind=kind, http_client=http_client)
    session.start()
    session.wait_for_idle = MagicMock(side_effect=ValueError("yo"))
    command = Command("command", spark_events=spark_events)

    try:
        result = command.execute(session)
        assert False
    except ValueError as e:
        spark_events.emit_statement_execution_start_event.assert_called_with(
            session.guid, session.kind, session.id, command.guid
        )
        spark_events.emit_statement_execution_end_event.assert_called_once_with(
            session.guid,
            session.kind,
            session.id,
            command.guid,
            -1,
            False,
            "ValueError",
            "yo",
        )
        assert_equals(e, session.wait_for_idle.side_effect)


@with_setup(_setup)
def test_execute_failure_post_statement_emits_event():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.get_statement.return_value = tls.TestLivySession.ready_statement_json
    session = _create_session(kind=kind, http_client=http_client)
    session.wait_for_idle = MagicMock()
    session.start()
    session.wait_for_idle = MagicMock()
    command = Command("command", spark_events=spark_events)

    http_client.post_statement.side_effect = KeyError("Something bad happened here")
    try:
        result = command.execute(session)
        assert False
    except KeyError as e:
        spark_events.emit_statement_execution_start_event.assert_called_once_with(
            session.guid, session.kind, session.id, command.guid
        )
        spark_events.emit_statement_execution_end_event._assert_called_once_with(
            session.guid,
            session.kind,
            session.id,
            command.guid,
            -1,
            False,
            "KeyError",
            "Something bad happened here",
        )
        assert_equals(e, http_client.post_statement.side_effect)


@with_setup(_setup)
def test_execute_failure_get_statement_output_emits_event():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.get_statement.return_value = tls.TestLivySession.ready_statement_json

    session = _create_session(kind=kind, http_client=http_client)
    session.wait_for_idle = MagicMock()
    session.start()
    session.wait_for_idle = MagicMock()
    command = Command("command", spark_events=spark_events)
    command._get_statement_output = MagicMock(side_effect=AttributeError("OHHHH"))

    try:
        result = command.execute(session)
        assert False
    except AttributeError as e:
        spark_events.emit_statement_execution_start_event.assert_called_once_with(
            session.guid, session.kind, session.id, command.guid
        )
        spark_events.emit_statement_execution_end_event._assert_called_once_with(
            session.guid,
            session.kind,
            session.id,
            command.guid,
            -1,
            False,
            "AttributeError",
            "OHHHH",
        )
        assert_equals(e, command._get_statement_output.side_effect)


@with_setup(_setup)
def test_execute_interrupted():
    spark_events = MagicMock()
    kind = SESSION_KIND_SPARK
    http_client = MagicMock()
    http_client.get_statement.return_value = tls.TestLivySession.ready_statement_json
    session = _create_session(kind=kind, http_client=http_client)
    session.wait_for_idle = MagicMock()
    session.start()
    session.wait_for_idle = MagicMock()
    command = Command("command", spark_events=spark_events)

    mock_ipython = MagicMock()
    mock_get_ipython = lambda: mock_ipython
    mock_ipython._showtraceback = mock_show_tb = MagicMock()
    sparkmagic.livyclientlib.exceptions.get_ipython = mock_get_ipython
    http_client.post_statement.side_effect = KeyboardInterrupt("")
    try:
        result = command.execute(session)
        assert False
    except KeyboardInterrupt as e:
        spark_events.emit_statement_execution_start_event.assert_called_once_with(
            session.guid, session.kind, session.id, command.guid
        )
        spark_events.emit_statement_execution_end_event._assert_called_once_with(
            session.guid,
            session.kind,
            session.id,
            command.guid,
            -1,
            False,
            "KeyboardInterrupt",
            "",
        )
        assert isinstance(e, SparkStatementCancelledException)
        assert_equals(str(e), COMMAND_INTERRUPTED_MSG)

        # Test patching _showtraceback()
        assert mock_ipython._showtraceback is SparkStatementCancelledException._show_tb

        with _capture_stderr() as stderr:
            mock_ipython._showtraceback(KeyError, "Dummy KeyError", MagicMock())
            mock_show_tb.assert_called_once()
            assert not stderr.getvalue()

        with _capture_stderr() as stderr:
            mock_ipython._showtraceback(
                SparkStatementCancelledException, COMMAND_INTERRUPTED_MSG, MagicMock()
            )
            mock_show_tb.assert_called_once()  # still once
            assert_equals(stderr.getvalue().strip(), COMMAND_INTERRUPTED_MSG)

    except:
        assert False
    else:
        assert False



from hdijupyterutils.constants import INSTANCE_ID, EVENT_NAME, TIMESTAMP
from hdijupyterutils.utils import get_instance_id, generate_uuid
from nose.tools import with_setup, raises
from mock import MagicMock

import sparkmagic.utils.constants as constants
from sparkmagic.utils.sparkevents import SparkEvents


def _setup():
    global spark_events, guid1, guid2, guid3, time_stamp

    spark_events = SparkEvents()
    spark_events.handler = MagicMock()
    spark_events.get_utc_date_time = MagicMock()
    spark_events._verify_language_ok = MagicMock()
    time_stamp = spark_events.get_utc_date_time()
    guid1 = generate_uuid()
    guid2 = generate_uuid()
    guid3 = generate_uuid()


def _teardown():
    pass


@with_setup(_setup, _teardown)
def test_emit_library_loaded_event():
    event_name = constants.LIBRARY_LOADED_EVENT
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
    ]

    spark_events.emit_library_loaded_event()

    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_cluster_change_event():
    status_code = 200

    event_name = constants.CLUSTER_CHANGE_EVENT
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.CLUSTER_DNS_NAME, guid1),
        (constants.STATUS_CODE, status_code),
        (constants.SUCCESS, True),
        (constants.ERROR_MESSAGE, None),
    ]

    spark_events.emit_cluster_change_event(guid1, status_code, True, None)

    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_session_creation_start_event():
    language = constants.SESSION_KIND_SPARK
    event_name = constants.SESSION_CREATION_START_EVENT
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
    ]

    spark_events.emit_session_creation_start_event(guid1, language)

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_session_creation_end_event():
    language = constants.SESSION_KIND_SPARK
    event_name = constants.SESSION_CREATION_END_EVENT
    status = constants.BUSY_SESSION_STATUS
    session_id = 0
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.STATUS, status),
        (constants.SUCCESS, True),
        (constants.EXCEPTION_TYPE, ""),
        (constants.EXCEPTION_MESSAGE, ""),
    ]

    spark_events.emit_session_creation_end_event(
        guid1, language, session_id, status, True, "", ""
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_session_deletion_start_event():
    language = constants.SESSION_KIND_SPARK
    event_name = constants.SESSION_DELETION_START_EVENT
    status = constants.BUSY_SESSION_STATUS
    session_id = 0
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.STATUS, status),
    ]

    spark_events.emit_session_deletion_start_event(guid1, language, session_id, status)

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_session_deletion_end_event():
    language = constants.SESSION_KIND_SPARK
    event_name = constants.SESSION_DELETION_END_EVENT
    status = constants.BUSY_SESSION_STATUS
    session_id = 0
    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.STATUS, status),
        (constants.SUCCESS, True),
        (constants.EXCEPTION_TYPE, ""),
        (constants.EXCEPTION_MESSAGE, ""),
    ]

    spark_events.emit_session_deletion_end_event(
        guid1, language, session_id, status, True, "", ""
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_statement_execution_start_event():
    language = constants.SESSION_KIND_PYSPARK
    session_id = 7
    event_name = constants.STATEMENT_EXECUTION_START_EVENT

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.STATEMENT_GUID, guid2),
    ]

    spark_events.emit_statement_execution_start_event(
        guid1, language, session_id, guid2
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_statement_execution_end_event():
    language = constants.SESSION_KIND_SPARK
    session_id = 7
    statement_id = 400
    event_name = constants.STATEMENT_EXECUTION_END_EVENT
    success = True
    exception_type = ""
    exception_message = "foo"

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.STATEMENT_GUID, guid2),
        (constants.STATEMENT_ID, statement_id),
        (constants.SUCCESS, success),
        (constants.EXCEPTION_TYPE, exception_type),
        (constants.EXCEPTION_MESSAGE, exception_message),
    ]

    spark_events.emit_statement_execution_end_event(
        guid1,
        language,
        session_id,
        guid2,
        statement_id,
        success,
        exception_type,
        exception_message,
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_sql_execution_start_event():
    event_name = constants.SQL_EXECUTION_START_EVENT
    session_id = 22
    language = constants.SESSION_KIND_SPARK
    samplemethod = "sample"
    maxrows = 12
    samplefraction = 0.5

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.SQL_GUID, guid2),
        (constants.SAMPLE_METHOD, samplemethod),
        (constants.MAX_ROWS, maxrows),
        (constants.SAMPLE_FRACTION, samplefraction),
    ]

    spark_events.emit_sql_execution_start_event(
        guid1, language, session_id, guid2, samplemethod, maxrows, samplefraction
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_sql_execution_end_event():
    event_name = constants.SQL_EXECUTION_END_EVENT
    session_id = 17
    language = constants.SESSION_KIND_SPARK
    success = False
    exception_type = "ValueError"
    exception_message = "You screwed up"

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.SESSION_GUID, guid1),
        (constants.LIVY_KIND, language),
        (constants.SESSION_ID, session_id),
        (constants.SQL_GUID, guid2),
        (constants.STATEMENT_GUID, guid3),
        (constants.SUCCESS, success),
        (constants.EXCEPTION_TYPE, exception_type),
        (constants.EXCEPTION_MESSAGE, exception_message),
    ]

    spark_events.emit_sql_execution_end_event(
        guid1,
        language,
        session_id,
        guid2,
        guid3,
        success,
        exception_type,
        exception_message,
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_magic_execution_start_event():
    event_name = constants.MAGIC_EXECUTION_START_EVENT
    magic_name = "sql"
    language = constants.SESSION_KIND_SPARKR

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.MAGIC_NAME, magic_name),
        (constants.LIVY_KIND, language),
        (constants.MAGIC_GUID, guid1),
    ]

    spark_events.emit_magic_execution_start_event(magic_name, language, guid1)

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


@with_setup(_setup, _teardown)
def test_emit_magic_execution_end_event():
    event_name = constants.MAGIC_EXECUTION_END_EVENT
    magic_name = "sql"
    language = constants.SESSION_KIND_SPARKR
    success = True
    exception_type = ""
    exception_message = ""

    kwargs_list = [
        (INSTANCE_ID, get_instance_id()),
        (EVENT_NAME, event_name),
        (TIMESTAMP, time_stamp),
        (constants.MAGIC_NAME, magic_name),
        (constants.LIVY_KIND, language),
        (constants.MAGIC_GUID, guid1),
        (constants.SUCCESS, success),
        (constants.EXCEPTION_TYPE, exception_type),
        (constants.EXCEPTION_MESSAGE, exception_message),
    ]

    spark_events.emit_magic_execution_end_event(
        magic_name, language, guid1, success, exception_type, exception_message
    )

    spark_events._verify_language_ok.assert_called_once_with(language)
    spark_events.get_utc_date_time.assert_called_with()
    spark_events.handler.handle_event.assert_called_once_with(kwargs_list)


def test_magic_verify_language_ok():
    for language in constants.SESSION_KINDS_SUPPORTED:
        SparkEvents()._verify_language_ok(language)


@raises(AssertionError)
def test_magic_verify_language_ok_error():
    SparkEvents()._verify_language_ok("NYARGLEBARGLE")



"""Base class for implementing an authentication provider for SparkMagic"""

from hdijupyterutils.ipywidgetfactory import IpyWidgetFactory
from sparkmagic.utils.constants import WIDGET_WIDTH


class Authenticator(object):
    """Base Authenticator for all Sparkmagic authentication providers."""

    def __init__(self, parsed_attributes=None):
        """Initializes the Authenticator with the attributes in the attributes
        parsed from a %spark magic command if applicable, or with default values
        otherwise.

        Args:
            self,
            parsed_attributes (IPython.core.magics.namespace): The namespace object that
            is created from parsing %spark magic command.
        """
        if parsed_attributes is not None:
            self.url = parsed_attributes.url
        else:
            self.url = "http://example.com/livy"
        self.widgets = self.get_widgets(WIDGET_WIDTH)

    def get_widgets(self, widget_width):
        """Creates and returns an address widget

        Args:
            widget_width (str): The width of all widgets to be created.

        Returns:
            Sequence[hdijupyterutils.ipywidgetfactory.IpyWidgetFactory]: list of widgets
        """
        ipywidget_factory = IpyWidgetFactory()

        self.address_widget = ipywidget_factory.get_text(
            description="Address:", value="http://example.com/livy", width=widget_width
        )
        widgets = [self.address_widget]
        return widgets

    def update_with_widget_values(self):
        """Updates url to be value in address widget."""
        self.url = self.address_widget.value

    def __call__(self, request):
        """subclasses should override"""
        return None

    def __eq__(self, other):
        if not isinstance(other, Authenticator):
            return False
        return self.url == other.url

    def __hash__(self):
        return hash((self.url, self.__class__.__name__))



# coding=utf-8
from mock import MagicMock, call
from nose.tools import with_setup, assert_equals, assert_false, assert_raises
from pandas.testing import assert_frame_equal

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.constants import LONG_RANDOM_VARIABLE_NAME, MIMETYPE_TEXT_PLAIN
from sparkmagic.livyclientlib.sparkstorecommand import SparkStoreCommand
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.exceptions import BadUserDataException


backup_conf_defaults = None


def _setup():
    global backup_conf_defaults
    backup_conf_defaults = {
        "samplemethod": conf.default_samplemethod(),
        "maxrows": conf.default_maxrows(),
        "samplefraction": conf.default_samplefraction(),
    }


def _teardown():
    conf.override_all(backup_conf_defaults)


@with_setup(_setup, _teardown)
def test_to_command_pyspark():
    variable_name = "var_name"
    sparkcommand = SparkStoreCommand(variable_name)
    sparkcommand._pyspark_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command("pyspark", variable_name)
    sparkcommand._pyspark_command.assert_called_with(variable_name)


@with_setup(_setup, _teardown)
def test_to_command_scala():
    variable_name = "var_name"
    sparkcommand = SparkStoreCommand(variable_name)
    sparkcommand._scala_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command("spark", variable_name)
    sparkcommand._scala_command.assert_called_with(variable_name)


@with_setup(_setup, _teardown)
def test_to_command_r():
    variable_name = "var_name"
    sparkcommand = SparkStoreCommand(variable_name)
    sparkcommand._r_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command("sparkr", variable_name)
    sparkcommand._r_command.assert_called_with(variable_name)


@with_setup(_setup, _teardown)
def test_to_command_invalid():
    variable_name = "var_name"
    sparkcommand = SparkStoreCommand(variable_name)
    assert_raises(
        BadUserDataException, sparkcommand.to_command, "invalid", variable_name
    )


@with_setup(_setup, _teardown)
def test_sparkstorecommand_initializes():
    variable_name = "var_name"
    samplemethod = "take"
    maxrows = 120
    samplefraction = 0.6
    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod, maxrows, samplefraction
    )
    assert_equals(sparkcommand.samplemethod, samplemethod)
    assert_equals(sparkcommand.maxrows, maxrows)
    assert_equals(sparkcommand.samplefraction, samplefraction)


@with_setup(_setup, _teardown)
def test_sparkstorecommand_loads_defaults():
    defaults = {
        conf.default_samplemethod.__name__: "sample",
        conf.default_maxrows.__name__: 419,
        conf.default_samplefraction.__name__: 0.99,
    }
    conf.override_all(defaults)
    variable_name = "var_name"
    sparkcommand = SparkStoreCommand(variable_name)
    assert_equals(
        sparkcommand.samplemethod, defaults[conf.default_samplemethod.__name__]
    )
    assert_equals(sparkcommand.maxrows, defaults[conf.default_maxrows.__name__])
    assert_equals(
        sparkcommand.samplefraction, defaults[conf.default_samplefraction.__name__]
    )


@with_setup(_setup, _teardown)
def test_pyspark_livy_sampling_options():
    variable_name = "var_name"

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=120)
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).take(120): print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=-1)
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).collect(): print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.25, maxrows=-1
    )
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).sample(False, 0.25).collect(): "
            "print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.33, maxrows=3234
    )
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).sample(False, 0.33).take(3234): "
            "print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.33, maxrows=3234
    )
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).sample(False, 0.33).take(3234): "
            "print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod=None, maxrows=100)
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).take(100): print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod=None, maxrows=100)
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).take(100): print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )


@with_setup(_setup, _teardown)
def test_scala_livy_sampling_options():
    variable_name = "abc"

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=100)
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command("{}.toJSON.take(100).foreach(println)".format(variable_name)),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=-1)
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command("{}.toJSON.collect.foreach(println)".format(variable_name)),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.25, maxrows=-1
    )
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command(
            "{}.toJSON.sample(false, 0.25).collect.foreach(println)".format(
                variable_name
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.33, maxrows=3234
    )
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command(
            "{}.toJSON.sample(false, 0.33).take(3234).foreach(println)".format(
                variable_name
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod=None, maxrows=100)
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command("{}.toJSON.take(100).foreach(println)".format(variable_name)),
    )


@with_setup(_setup, _teardown)
def test_r_livy_sampling_options():
    variable_name = "abc"

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=100)

    assert_equals(
        sparkcommand._r_command(variable_name),
        Command(
            "for ({} in (jsonlite::toJSON(take({},100)))) {{cat({})}}".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=-1)
    assert_equals(
        sparkcommand._r_command(variable_name),
        Command(
            "for ({} in (jsonlite::toJSON(collect({})))) {{cat({})}}".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.25, maxrows=-1
    )
    assert_equals(
        sparkcommand._r_command(variable_name),
        Command(
            "for ({} in (jsonlite::toJSON(collect(sample({}, FALSE, 0.25))))) {{cat({})}}".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(
        variable_name, samplemethod="sample", samplefraction=0.33, maxrows=3234
    )
    assert_equals(
        sparkcommand._r_command(variable_name),
        Command(
            "for ({} in (jsonlite::toJSON(take(sample({}, FALSE, 0.33),3234)))) {{cat({})}}".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )

    sparkcommand = SparkStoreCommand(variable_name, samplemethod=None, maxrows=100)
    assert_equals(
        sparkcommand._r_command(variable_name),
        Command(
            "for ({} in (jsonlite::toJSON(take({},100)))) {{cat({})}}".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )


@with_setup(_setup, _teardown)
def test_execute_code():
    spark_events = MagicMock()
    variable_name = "abc"

    sparkcommand = SparkStoreCommand(
        variable_name, "take", 100, 0.2, spark_events=spark_events
    )
    sparkcommand.to_command = MagicMock(return_value=MagicMock())
    result = """{"z":100, "nullv":null, "y":50}
{"z":25, "nullv":null, "y":10}"""
    sparkcommand.to_command.return_value.execute = MagicMock(
        return_value=(True, result, MIMETYPE_TEXT_PLAIN)
    )
    session = MagicMock()
    session.kind = "pyspark"
    result = sparkcommand.execute(session)

    sparkcommand.to_command.assert_called_once_with(session.kind, variable_name)
    sparkcommand.to_command.return_value.execute.assert_called_once_with(session)


@with_setup(_setup, _teardown)
def test_unicode():
    variable_name = "collect 'è'"

    sparkcommand = SparkStoreCommand(variable_name, samplemethod="take", maxrows=120)
    assert_equals(
        sparkcommand._pyspark_command(variable_name),
        Command(
            "import sys\nfor {} in {}.toJSON(use_unicode=(sys.version_info.major > 2)).take(120): print({})".format(
                LONG_RANDOM_VARIABLE_NAME, variable_name, LONG_RANDOM_VARIABLE_NAME
            )
        ),
    )
    assert_equals(
        sparkcommand._scala_command(variable_name),
        Command("{}.toJSON.take(120).foreach(println)".format(variable_name)),
    )
